/* -*- Mode: JavaScript; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * Copyright (c) 2006, 2007 ActiveState Software Inc.
 * Refer to the license in the Komodo IDE installation
 * for the licensing information of this source code.
 *
 * winger-commands.js -- debugger commands
 */

if (typeof(JSDBGP) == 'undefined') {
    var JSDBGP = {};
}

JSDBGP.Command = function(mainObj) {
    this.mainObj = mainObj;
    this.notifications = [];
    this.stopReasons = ['starting', 'stopping', 'stopped', 'running', 'break', 'interactive'];
    this.stopReason = JSDBGP.Constants.STOP_REASON_STARTING;
    this._motion_commands = {
        'break' : null,
        'detach' : null,
        'step_into' : null,
        'step_out' : null,
        'step_over' : null,
        'stop' : null,
        '__end__' : null
    };
    this._htmlEventHandlerNames = {
        'onblur' : null,
        'onchange' : null,
        'onclick' : null,
        'ondblclick' : null,
        'onfocus' : null,
        'onkeydown' : null,
        'onkeypress' : null,
        'onkeyup' : null,
        'onload' : null,
        'onmousedown' : null,
        'onmousemove' : null,
        'onmouseout' : null,
        'onmouseover' : null,
        'onmouseup' : null,
        'onreset' : null,
        'onresize' : null,
        'onselect' : null,
        'onsubmit' : null,
        'onunload' : null,
        '__end__' : null
    };
    this.javascript_urls = {};
    this.dbgp_eval_ns = "x-dbgp://eval/x";

    this.log = JSDBGP.Logging.getLogger("winger.commands");
    this.log.setLevel(JSDBGP.Logging.DEBUG);
    this.breakClosure = null;
    this.error_code = null;

    this.ibState = JSDBGP.Constants.IB_STATE_NONE;
    this.ibBuffer = "";
    this.ibInBackslashContinuation = false; //XXX shouldn't be necessary
    this.prompts = [">", "*"]
    this._all_white_space_re = /^\s+$/;
    this._missing_closer_re = /^missing [\}\)\]] /;

    this._knownScriptHandlers = {};

    this.finalInit = function() {
        this.console = this.mainObj.console;
        this.propInfo = this.mainObj.properties;
        this.settings = this.mainObj.settings;
        this.utils = this.mainObj.utils;
    };

    this.doEnterNestedEventLoop = function(stopReason) {
        var console = this.console;
        console.nested_event_loop_depth += 1;
        this.stopReason = arguments.length >= 1 ? stopReason : JSDBGP.Constants.STOP_REASON_BREAK;



        var self = this;
        JSDBGP.ui_set_status(JSDBGP.Constants.DBGP_STATE_BREAK);
        console.check_for_run_end_only = false;
        var new_depth = console.jsds.enterNestedEventLoop({onNest: function() {



            self.console.session.make_input_pump();



        }});



    };

    this.doExitNestedEventLoop = function() {
        this.lastContinuationCommand = this.cmd;
        this.lastContinuationStatus = "break";
        this.lastTranID = this.transaction_id;
        JSDBGP.ui_set_status(JSDBGP.Constants.DBGP_STATE_RUN);
        this.console.session.notifyLeavingNestedEventLoop();
        try {
            if (this.console.nested_event_loop_depth >= 1) {
                this.console.nested_event_loop_depth -= 1;



                var new_depth = this.console.jsds.exitNestedEventLoop();



            } else {



            }
        } catch(ex) {
            this.log.exception(ex);



        }
        this.mainObj.debuggerObj.processHooksOffScripts();
    };

    this.re_get_cmd = /^\s*(\w+)\s*(.*)\s*$/;
    this.re_get_cmd._is_internal = true;
    this.re_context_get = /^\s*(\w+)[\s:]*(\d*)\s*$/;
    this.re_context_get._is_internal = true;
    this.re_jsExt = /\.js$/;
    this.re_baseHandlerName = /^on[a-z]+$/;
    this._cached_script_handler_coords = {};
    
    this._re_get_arg_1 = /^\s*(-\w)\s*([^\s\"\'\\]+)\s*(.*)$/;
    this._re_get_arg_2 = /^\s*(-\w)\s*"((?:\\.|[^\\\"])+)"\s*(.*)$/;
    this._re_get_arg_3 = /^\s*(--)\s+(.*?)\s*$/;

    this._parse_args = function(raw_args) {
        var args = {};
        var reo;
        while (raw_args.length > 0) {
            reo = this._re_get_arg_1.exec(raw_args);
            if (reo) {



                args[reo[1]] = reo[2];
                raw_args = reo[3];
            } else if ((reo = this._re_get_arg_2.exec(raw_args))) {



                args[reo[1]] = this.utils.cleanString(reo[2]);
                raw_args = reo[3];
            } else if ((reo = this._re_get_arg_3.exec(raw_args))) {



                args.data = reo[2];
                break;
            } else {
                args.data = raw_args;
                break;
            }
        }
        return args;
    };
    this._parse_args._is_internal = true;

    
    
    /* info: array of 
     *    [key ("-x") , objName, type:i|b, defaultValue]
     *    [key ("-y") , ... ]
     *    ["data", <dataName>]
     *
     */
    this._args_translate = function(args, args_obj, info) {
        for (var i in info) {
            var argItem = info[i];
            if (argItem[0] in args) {
                var argValue = args[argItem[0]];
                if (argItem[2]) {
                    var argType = argItem[2];
                    if (argType == "i") {



                        argValue = parseInt(argValue);



                    } else if (argType == "b") {  //boolean



                        if (/^\d+$/.test(argValue)) {
                            argValue = parseInt(argValue);
                        }
                        argValue = (argValue ? true : false);



                    } else {
                        this.error_code = JSDBGP.Constants.E_PARSE_ERROR;
                        throw new Error("Unexpected arg type of " + argType);
                    }
                }
                args_obj[argItem[1]] = argValue;



            } else if (typeof(argItem[3]) != "undefined") {
                args_obj[argItem[1]] = argItem[3];



            } else {



            }
        }
    };
    this._args_translate._is_internal = true;

    this._args_message = function(specificMsg, args) {
        return (specificMsg + " (args <" + args + ">)");
    };
    this._args_message._is_internal = true;
    

    // If a command returns true, it means exit the get-command loop
    this.dispatch = function(full_cmd) {



        if (!this.console.initialized) {
            this.log.debug("dispatch: debugger has shut down");
            return false;
        }
        var res;
        var disable_debugger = false;
        try {
            if (!full_cmd) {
                this.error_code = JSDBGP.Constants.E_PARSE_ERROR;
                throw new Error("No command given");
            }
            full_cmd = full_cmd.replace(/\r\n|\r|\n/g, '');
            var reo = this.re_get_cmd.exec(full_cmd);
            if (!reo) {
                this.error_code = JSDBGP.Constants.E_PARSE_ERROR;
                throw new Error("Can't get a command out of [" + full_cmd + "]");
            }
            var cmd = reo[1];
            // This might depend on the command we're running.
            disable_debugger = !(cmd in this._motion_commands);
            if (disable_debugger) {
                this.mainObj.debuggerObj.pushHooksOff();
            }
            if (!(cmd in this)) {
                this.error_code = JSDBGP.Constants.E_UNIMPLEMENTED_CMD;
                throw new Error("Don't know how to handle command [" + cmd +
                      "] in <<" + full_cmd + ">>");
            } else if (this[cmd]._is_internal) {
                this.error_code = JSDBGP.Constants.E_UNIMPLEMENTED_CMD;
                throw new Error("Won't dispatch to internal function " + cmd);
            }
            var args = this._parse_args(reo[2] || "");
            this.cmd = cmd;
            if ('-i' in args) {
                this.transaction_id = args['-i'];
            } else {
                this.transaction_id = this.lastTranID;
            }
            delete args['-i'];
            if ('-z' in args) {
                this.thread_id = args['-z'];
                delete args['-z'];
            } else {
                delete this.thread_id;
            }
            res = this[cmd](args, cmd);
        } catch(ex) {
            this.log.exception(ex);
            var response = this._get_xml_response_start_1();
            response.error.@code = this.error_code ? this.error_code[0] : JSDBGP.Constants.E_UNKNOWN[0];
            response.error.@apperr = this.error_code ? this.error_code[1] : JSDBGP.Constants.E_UNKNOWN[1];
            this.error_code = null;
            var msg = ex.toString();
            if (msg.indexOf(JSDBGP.Constants.E_MSG_HEADER) == -1) {
                msg = JSDBGP.Constants.E_MSG_HEADER + msg;
            }
            response.error.message = msg;
            this._addHeader_WriteXMLObject(response);
            res = false;
        }
        if (this.notifications) {
            // Schedule the tasks
            for (var i in this.notifications) {
                var func = this.notifications[i];
                try {
                    func();
                } catch(ex) {
                    this.log.exception(ex, "can't run func [" + i + "] : " + func);
                }
            }
            this.notifications = [];
        }
        if (disable_debugger) {
            this.mainObj.debuggerObj.restoreHooks();
        }




        return res;
    };
    this.dispatch._is_internal = true;

    this.fallback = function(ex) {
        var response = this._get_xml_response_start_1();
        var ecode = JSDBGP.Constants.E_INTERNAL_ERROR;
        var error = <error />;
        error.@code = ecode[0];
        error.@apperr = ecode[1];
        error.message = ex.toString();;
        response.error = error;
        this._addHeader_WriteXMLObject(response);
    };
    this.fallback._is_internal = true;

    this._addHeader_WriteXMLObject = function(xmlObj) {
        if (!this.console.session) {
            this.log.debug("_addHeader_WriteXMLObject: debugger session is closed");
        } else {
            var objStr = typeof(xmlObj) == "xml" ? xmlObj.toXMLString() : xmlObj;
            this.console.session.writeData(this.xmlHeader + "\n" + objStr);
        }
    };
    
    // Breakpoint command handlers.

    this.breakpoint_set = function(args) {
            
        function x(args) {
            var s = "";
            for (var a in args) {
                s += (a + " " + args[a] + " ");
            }
            return s;
        }
        var args_obj = {};
        this._args_translate(args, args_obj,
                             [['-s', 'state', null, "enabled"],
                              ['-t', 'type'],
                              ['-f', 'filename'],
                              ['-n', 'lineno', 'i'],
                              ['-m', 'function'],
                              ['-x', 'exception'],
                              ['-h', 'hit_value', 'i', 0],
                              ['-o', 'hit_condition', null, ">="],
                              ['-r', 'is_temporary', 'b', false],
                              ['-z', 'threadID'],
                              ['data', 'expression']]);
        
        if ('filename' in args_obj && args_obj.filename.indexOf("dbgp:///") == 0) {
            args_obj.filename = args_obj.filename.substr("dbgp:///".length);
        }        
        switch (args_obj.type) {
        case 'line':
            if (!('filename' in args_obj && 'lineno' in args_obj)) {
                this.error_code = JSDBGP.Constants.E_BREAKPOINT_INVALID;
                throw new Error(this._args_message("Breakpoint type of line given, but both filename (-f) and line# (-n) aren't given", x(args)));
            }
            break;
            
        case 'call':
        case 'return':
            if (!('function' in args_obj)) {
                this.error_code = JSDBGP.Constants.E_BREAKPOINT_TYPE_NOT_SUPPORTED;
                throw new Error(this._args_message("Breakpoint type of function given, but no function (-m) specified", x(args)));
            }
            break;
            
        case 'exception':
            if (!('exception' in args_obj)) {
                this.error_code = JSDBGP.Constants.E_BREAKPOINT_TYPE_NOT_SUPPORTED;
                throw new Error(this._args_message("Breakpoint type of exception given, but no exceptionName (-x) specified", x(args)));
            }
            break;
            
        case 'conditional':
        case 'watch':
            if (!('expression' in args_obj)) {
                this.error_code = JSDBGP.Constants.E_BREAKPOINT_TYPE_NOT_SUPPORTED;
                throw new Error(this._args_message("Breakpoint type of " + args_obj.type + " given, but no expression specified", x(args)));
            }
            break;
            
        default:
            this.error_code = JSDBGP.Constants.E_BREAKPOINT_INVALID;
            throw new Error("Unknown breakpoint type of " + args_obj.type);
        }
        if (args_obj.type != 'line') {
            this.error_code = JSDBGP.Constants.E_BREAKPOINT_TYPE_NOT_SUPPORTED;
            throw new Error(this._args_message("Not yet supporting breakpoints of type " + args_obj.type, x(args)));
        }            

        if (!(args_obj.state == "enabled" || args_obj.state == "disabled")) {
            this.error_code = JSDBGP.Constants.E_BREAKPOINT_INVALID_STATE;
            throw new Error(this._args_message("Invalid breakpoint state of " + args_obj.state, x(args)));
        }

        if (args_obj.hit_value && !(args_obj.hit_condition in {">=":null, "==":null, "%":null})) {
            this.error_code = JSDBGP.Constants.E_BREAKPOINT_INVALID;
            throw new Error(this._args_message("invalid hit_condition of " + args_obj.hit_condition, x(args)));
        }
        var fileName_raw = this._fix_uri(args_obj.filename);
        args_obj.filename = this.mainObj.scripts.mapFileURI_to_HTTP_URI(fileName_raw);

        var bkpt_id = this.mainObj.breakpointManager.addFileBreakpoint(args_obj);
        var response = this._get_xml_response_start_1();
        response.@state = args_obj.state;
        response.@id = bkpt_id;
        this._addHeader_WriteXMLObject(response);
        return false;
    };

    this._get_breakpoint_id = function(args) {
        var args_obj = {};
        this._args_translate(args, args_obj,
                             [['-d', 'breakpoint_id']]);
        if (!('breakpoint_id' in args_obj)) {
            this.error_code = JSDBGP.Constants.E_BREAKPOINT_INVALID;
            throw new Error(this._args_message("breakpoint_get/list: no breakpoint id (-d) specified", x(args)));
        }
        return args_obj.breakpoint_id;
    };
    this._get_breakpoint_id._is_internal = true;

    this.breakpoint_get = function(args) {
        var bkpt_id = this._get_breakpoint_id(args);
        var bkpt = this.mainObj.breakpointManager.getBreakpoint(args_obj.breakpoint_id);
        if (!bkpt) {
            this.error_code = JSDBGP.Constants.E_BREAKPOINT_NO_ID;
            throw new Error("BreakpointManager.getBreakpoint: breakpoint ID " + bkpt_id + " not found");
        }
        var response = this._get_xml_response_start_1();
        var bkpt_tag = bkpt.display_e4x();
        response.breakpoint = bkpt_tag; //e4x
        this._addHeader_WriteXMLObject(response);
        return false;
    };

    this.breakpoint_list = function(args) {
        var breakpointManager = this.mainObj.breakpointManager;
        var response = this._get_xml_response_start_1();
        var bkpt_list = this.mainObj.breakpointManager.getAllBreakpoints();
        for (var i in bkpt_list) {
            var bkpt = bkpt_list[i];
            response.breakpoint += bkpt.display_e4x();
        }
        this._addHeader_WriteXMLObject(response);
        return false;
    };

    this.breakpoint_remove = function(args) {
        var bkpt_id = this._get_breakpoint_id(args);
        this.mainObj.breakpointManager.removeBreakpoint(bkpt_id);
        var response = this._get_xml_response_start_1();
        this._addHeader_WriteXMLObject(response);
        return false;
    };

    /*
     *
     * breakpoint_update:
     *     -d bkpt_id
     *     [-s state=a|b|c : <current state>]
     *     [-n lineno : <current line>]
     *     [-h hit_value : <current hit_value>]
     *     [-o hit_condition : <current hit_condition>]
     */
    this.breakpoint_update = function(args) {
        var args_obj = {};
        this._args_translate(args, args_obj,
                             [['-d', 'breakpoint_id'],
                              ['-s', 'state'],
                              ['-n', 'lineno', 'i'],
                              ['-h', 'hit_value', 'i', 0],
                              ['-o', 'hit_condition'],
                              ['data', 'expression']]);
        if (!('breakpoint_id' in args_obj)) {
            this.error_code = JSDBGP.Constants.E_BREAKPOINT_INVALID;
            throw new Error(this._args_message("breakpoint_update: no breakpoint id (-d) specified", x(args)));
        }
        // Make sure on of the args was given
        var arg_count = 0;
        for (var tag in ['state', 'lineno', 'hit_value', 'hit_condition', 'expression']) {
            ++arg_count;
        }
        if (arg_count > 0) {
            this.mainObj.breakpointManager.updateBreakpoint(args_obj.breakpoint_id, args_obj);
        } else {



        }
        var response = this._get_xml_response_start_1();
        this._addHeader_WriteXMLObject(response);
        return false;
    };

    this.feature_get = function(args) {
        var settings = this.settings;
        var featureName = args['-n'];
        var innerText = "";
        var supported;



        if (typeof(featureName) == "undefined") {
            featureName = "unspecified";
            supported = 0;
        } else if (featureName in settings.supportedCommands) {
            supported = settings.supportedCommands[featureName];
        } else if (featureName in settings.supportedFeatures) {
            var vals = settings.supportedFeatures[featureName];
            supported = vals[0];
            if (featureName in settings.registeredFunctions) {
                innerText = settings.get(featureName);
            } else if (vals[2] == 0 || !(featureName in settings.settings)) {
                innerText = "";
            } else {
                innerText = settings.settings[featureName][0];
            }
        } else {
            // Command not recognized
            supported = 0;
        }
        var response = this._get_xml_response_start_1();
        response.@feature_name = featureName;
        response.@supported = supported;
        response.appendChild(innerText);
        this._addHeader_WriteXMLObject(response);
        return false;
    };

    this.feature_set = function(args) {
        var settings = this.settings;
        var featureName = args['-n'];
        var featureValue = args['-v'];
        if (/^-?\d+$/.test(featureValue)) {
            featureValue = parseInt(featureValue);
        }
        var reason = null;
        var status = 0;
        var success;
        if (typeof(featureName) == "undefined" || !featureName) {
            success = 0;
            reason = "Command not specified";
        } else if (!(featureName in settings.supportedFeatures)) {
            status = 0;
            reason = "Command " + featureName + " not recognized";
        } else {
            var vals = settings.supportedFeatures[featureName];
            if (vals[1] == 0) {
                if (featureName in settings.registeredFunctions) {
                    status = settings.set(featureName, featureValue);
                } else {
                    status = 0;
                    reason = "Command " + featureName + " not modifiable";
                }
            } else if (vals[2] == 0) {
                // No associated data, use boolean value in table
                vals[0] = featureValue ? 1 : 0;
                status = 1;
                success = vals[0];
            } else if (!(featureName in settings.settings)) {
                status = 0;
                reason = "Command " + featureName + " not in settings table";
            } else {
                var svals = settings.settings[featureName][1];
                if (typeof(svals) == "undefined" || svals == null) {
                    status = 0;
                    reason ="Command " + featureName + " is readonly settings table";
                } else if (this.propInfo.isArray(svals)) {
                    status = 0;
                    for (var i = 0; i < svals.length; i++) {
                        var allowedValue = svals[i];
                        if (featureValue == allowedValue) {
                            status = 1;
                            settings.settings[featureName][0] = featureValue;
                            if (status == 1 && featureName == 'data_encoding') {
                                this.propInfo.default_encoding = featureValue;
                            }
                            break;
                        }
                    }
                    if (status == 0) {
                        reason = "Command " + featureName + " value of " + featureValue + " isn't an allowed value.";
                    }
                } else if (svals == 1) {
                    // Assign numeric values
                    if (/^-?\d+/.test(featureValue)) {
                        status = 1;
                        settings.settings[featureName][0] = featureValue;
                    } else {
                        status = 0;
                        reason = "Command " + featureName + " value of " + featureValue + " isn't numeric.";
                    }
                } else if (svals == 'a') {
                    // Allow any ascii data
                    status = 1;
                    settings.settings[featureName][0] = featureValue;
                } else {
                    status = 0;
                    reason = "Command " + featureName + "=" + featureValue + ", can't deal with current setting of " + pp(vals) + "\n";
                }
            }
        }
        var response = this._get_xml_response_start_1();
        response.@feature_name = featureName;
        response.@success = status;
        if (reason) {
            response.@reason  = reason; // xml-encoding done by e4x
        }
        this._addHeader_WriteXMLObject(response);
        return false;
    };

    // Helpers for context_get:

    // Most recent frame is at position 0, due to the way
    // the frame-stack is built each time.
    
    this._get_non_native_stack_frame = function(frames, stackDepth) {
        if (stackDepth > frames.length) {
            stackDepth = frames.length - 1;
        }
        // Stop when we hit -1
        var j = stackDepth;
        for (var i = stackDepth; i >= 0; --i) {
            var f = frames[stackDepth];
            if (f.isNative) {
                continue;
            }
            var jsdScript = f.script;
            if (!jsdScript) {
                // This frame is native, so skip it
                this.log.error("_get_non_native_stack_frame: stack frame " + i + "isn't native, but has no script!");
                continue;
            }
            if (j == 0) {
                return f;
            }
            --j;
        }
        this.error_code = JSDBGP.Constants.E_STACK_DEPTH_INVALID;
        throw new Error("_get_non_native_stack_frame: can't find non-native stack " + stackDepth);
    };

    this._merge_rest = function(aret, srca, ax) {
        var lim = srca.length;
        for (; ax < lim; ++ax) {
            aret.push(srca[ax]);
        }
        return aret;
    };
    this._merge_locals = function(resa, resb) {
	var alim = resa.length;
	var blim = resb.length;
	var ax = 0, bx = 0;
	var aret= [];
	while (ax < alim && bx < blim) {
            if (resa[ax].name == resb[bx].name) {
                // the old one takes precedence over the new one,
                // as older names are more local
		aret.push(resa[ax++]);
                ++bx;
	    } else if (resa[ax].name < resb[bx].name) {
		aret.push(resa[ax++]);
	    } else {
		aret.push(resb[bx++]);
	    }
	}
	if (ax < alim) {
	    return this._merge_rest(aret, resa, ax);
	} else if (bx < blim) {
	    return this._merge_rest(aret, resb, bx);
	} else {
            return aret;
        }
    };

    this._buildPropertyList = function(context_type, response, resObj, hide_hidden) {

        var currVal = resObj.pivot_value;
        var curr_frame = resObj.curr_frame;

        // set max_num_to_return to 100 to prevent timeouts for now.
        // here currVal is always a jsval wrapped as a jsdIValue
        var resa = this.propInfo.walkProperties(currVal, true, hide_hidden, false, 100); //XXX Should be: this.settings.settings['max_children'][0]);
        if (context_type == this.propInfo.LocalVars && curr_frame && curr_frame.isValid) {
            var globalObject = curr_frame.executionContext.globalObject;
	    var scope_num = 0;
            var parent;
	    while (true) {
		parent = currVal.jsParent;
                if (!parent) {



                    break;
                } else if (!parent.jsParent || parent.jsParent.jsType == JSDBGP.Constants.TYPE_NULL) {



                    break;
                } else if (parent == globalObject) {



		    break;
                }
		scope_num++;
		var res_new = this.propInfo.walkProperties(parent, true, hide_hidden, false, 100);
		resa = this._merge_locals(resa, res_new);
                currVal = parent;
	    }
	}
		
        var divisor = resa.length;
        if (divisor == 0) divisor = 1;
        var max_data_size = Math.round(this.settings.settings['max_data'][0] / divisor);
        for (var i = 0; i < resa.length; i++) {
            response.property += this.propInfo.expandTopLevelProperty(resa[i], hide_hidden, max_data_size);
        }
    };

    // returns the frame and the pivot-value
    // Precondition: all fields in resObj have been initialized to null
    this._getFrameInfo = function(context, context_depth, resObj) {
        if (!this.console.frames) {
            if (context == this.propInfo.GlobalVars) {
                resObj.pivot_value = this.console.jsds.wrapValue(window);
            }
            return;
        }
        resObj.curr_frame = this._get_non_native_stack_frame(this.console.frames, context_depth);
        if (resObj.curr_frame && !resObj.curr_frame.isValid) {
            resObj.curr_frame = resObj.pivot_value = null;
            return;
        }
        switch (context) {
        case this.propInfo.LocalVars:
            resObj.pivot_value = resObj.curr_frame.scope;
            break;
            
        case this.propInfo.GlobalVars:
            resObj.pivot_value = resObj.curr_frame.executionContext.globalObject;
            break;
            
        case this.propInfo.InstanceVars:
            resObj.pivot_value = resObj.curr_frame.thisValue;
            break;

        default:
            this.error_code = JSDBGP.Constants.E_CONTEXT_INVALID;
            throw new Error("Expected context_get globals|locals|this, got [" + context + "]");
        }
    };

    /* Syntax:
     * context_get type=(locals|object|globals) [depth=0]
     * stack-depth = 0 : current frame.
     * writes back a list of [name, type, value, #children]
     */
    this.context_get = function(args) {
        var context_type = parseInt(this._getArg(args, '-c', this.propInfo.LocalVars));
        var context_depth = parseInt(this._getArg(args, '-d', "0"));
        var response = this._get_xml_response_start_1();
        var resObj = {curr_frame:null, pivot_value:null};
        this._getFrameInfo(context_type, context_depth, resObj);
        if (resObj.pivot_value) {
            this._buildPropertyList(context_type, response, resObj,
                                    !this.settings.settings['show_hidden'][0]);
            if (resObj.curr_frame) {
                this.setWalkedFrame(resObj.curr_frame, context_type);
            }
        }
        this._addHeader_WriteXMLObject(response);
        return false;
    };

    this.context_names = function(args) {
        try {
            var nlist = this.propInfo.contextPropertyNames;
            var response = this._get_xml_response_start_1();
            for (var i = 0; i < nlist.length; i++) {
                var cname = nlist[i];
                var cval = this.propInfo.contextProperties[cname];
                response.context += <context name={cname} id={cval} />
                                                              }
            this._addHeader_WriteXMLObject(response);
        } catch(ex) {
            this.log.exception(ex, "Failure trying to get context_names");
            response = this._get_xml_response_start_1();
            response.error = <error code="404" />;
            response.error.message = ex.toString();
            this._addHeader_WriteXMLObject(response);
        }
        return false;
    };

    // "break" is a reserved js word
    // JS reserves keywords everywhere, even after dots.
    this.doBreakNextStatement = function(closure) {
        this.breakClosure = closure;
        this.doBreak();
    };
    
    this.doBreak = function(args) {
        var debuggerObj = this.mainObj.debuggerObj;
        debuggerObj.setStopState(true);
        debuggerObj.clearStepPast();
        return true;
    };
    // JavaScript doesn't allow keywords after dots, so this comment
    // is here for searching purposes only
    // this.break = function(args) {...}
    this['break'] = this.doBreak;

    this.detach = this.stop;

    this._doEval = function(buf) {
        // Try to eval the expn in the top frame, but if it
        // fails do an all-encompassing eval().

        if (!this.console.frames || this.console.frames.length == 0) {
            return this._mochiEval(buf);
        }
        var f = this._get_non_native_stack_frame(this.console.frames, 0);
        var resObj = {};
        var bres = f.eval(buf, this.dbgp_eval_ns, 1, resObj);
        if (!bres || !resObj.value) {
            // Let raw eval create an exception that the caller handles
            // This is how incomplete blocks are handled.
            return this._mochiEval(buf);
        }
        var wrappedValue = resObj.value.getWrappedValue();
        // Put functions in the global scope also
        // Also put assignments in there.
        if (typeof(wrappedValue) == "function"
            || buf.indexOf("function") == 0
            || buf.match(/[^=]=[^=]/)) {
            try {
                this._mochiEval(buf);
            } catch(ex) {
            }
        }
        return wrappedValue;
    };

    // Avoiding adding names taken from MochiKit, under the MIT License,
    // but it doesn't seem to matter.
    this._mochiEval = function () {
        var self = this;
        // Protect the 'this'
        var innerObj = {
            __func: function() {
                if (self.console.lastWindow) {
                    // Address bug 48798
                    // See discussion at https://bugzilla.mozilla.org/show_bug.cgi?id=293933
                    // and uses the scope-chain mechanism on eval
                    // where the second arg can be a window object.
                    return eval(arguments[0], self.console.lastWindow);
                }
                if (window) {
                    if (typeof(eval.call) != "undefined") {
                        return eval.call(window, arguments[0]);
                    } else {
                        with (window) { __ = eval(arguments[0]); }
                        return __;
                    }
                } else {
                    return eval(arguments[0]);
                }
            }
        };
        return innerObj.__func(arguments[0]);
    };

    // This is used to address bug 48798:
    // maintain user context when they aren't stopped at a breakpoint.
    
    this._interact_saveWindow = function() {
        var f = this._get_non_native_stack_frame(this.console.frames, 0);
        var holdCurrVal = {};
        var bres = f.eval("window", this.dbgp_eval_ns, 1, holdCurrVal);
        if (bres) {
            this.console.lastWindow = holdCurrVal.value.getWrappedValue();
            return;
        }
        this.console.lastWindow = null;



    };

    this._doOutput = function(streamName, msg) {
        var response = this._get_xml_output(streamName);
        if (msg.length == 0) {
            return;
        }
        if (msg[msg.length - 1] != "\n") {
            // Tack on a newline
            msg += "\n";
        }
        var encval;
        if (this.settings.settings['data_encoding'][0] == "base64") {
            encval = this.mainObj.utils.base64Encode(msg);
        } else {
            encval = msg;
        }
        response.appendChild(encval);
        this._addHeader_WriteXMLObject(response);
    };

    this._doStderr = function(msg) {
        return this._doOutput('stderr', msg);
    };

    this._doStdout = function(msg) {
        return this._doOutput('stdout', msg);
    };

    this.interact = function(args) {
        var args_obj = {};
        this._args_translate(args, args_obj,
                             [['-a', 'abort'],
                              ['-m', 'mode', 'i', -1]]);
        var do_abort = 'abort' in args_obj ? args_obj.abort : false;
        var i_mode = args_obj.mode;
        var response;
        if (this.ibState == JSDBGP.Constants.IB_STATE_NONE) {
            this.ibState = JSDBGP.Constants.IB_STATE_START;
            this.ibInBackslashContinuation = false;
            if (this.console.frames && this.console.frames.length > 0) {
                this._interact_saveWindow();
            }
        }
        if (i_mode == 0) {
            this.ibState = JSDBGP.Constants.IB_STATE_NONE;
            this.stopReason = JSDBGP.Constants.STOP_REASON_BREAK;
            response = this._get_xml_response_start_1();
            response.@status = this._getStopReason();
            response.@reason = "ok";
            response.@more = 0;
            response.@prompt = "";
            this._addHeader_WriteXMLObject(response);
            return;
        }
        this.stopReason = JSDBGP.Constants.STOP_REASON_INTERACT;
        var moreValue = 0;
        if (!do_abort && args.data) {
            var decodedData = this.utils.base64Decode(args.data);
            // If we're continuing, put a newline between the two lines,
            // but watch out for unterminated string errors
            if (this.ibState == JSDBGP.Constants.IB_STATE_START) {
                // Strip out all white-space
                this.ibBuffer = decodedData.replace(this._all_white_space_re, '');
                this.ibInBackslashContinuation = false;
            } else {
                if (this.ibInBackslashContinuation) {
                    this.ibInBackslashContinuation = false;
                } else {
                    // Add a newline for recognized multi-line constructs only
                    this.ibBuffer += "\n"
                }
                this.ibBuffer += decodedData;
            }
            if (this.ibBuffer.length > 0) {
                var bufLenSub1 = this.ibBuffer.length - 1;
                // A backslash at the end calls for continuation
                if (this.ibBuffer[bufLenSub1] == '\\') {
                    this.ibBuffer = this.ibBuffer.slice(0, -1);
                    this.ibState = JSDBGP.Constants.IB_STATE_PENDING;
                    this.ibInBackslashContinuation = true;
                    moreValue = 1;








                } else {
                    var res;
                    var mainError = null;
                    try {
                        // This happens on copy & paste
                        this.ibBuffer = this.ibBuffer.replace(/\\\n/mg, '');




                        res = this._doEval(this.ibBuffer);
                        this.ibState = JSDBGP.Constants.IB_STATE_START;
                    } catch(ex) {



                        res = null;
                        var doContinue;
                        if (ex instanceof SyntaxError || ex.name == "SyntaxError") {
                            //@@@@ RE
                            if (ex.message.match(this._missing_closer_re)
                                || ex.message.indexOf("unexpected end of XML source") == 0) {
                                // missing } in compound statement
                                // missing } after function body
                                // missing ] after element list
                                doContinue = true;
                            } else {
                                doContinue = false;
                            }
                        } else {
                            doContinue = false;
                        }
                        if (doContinue) {
                            moreValue = 1;
                            mainError = null;
                            this.ibState = JSDBGP.Constants.IB_STATE_PENDING;
                        } else {
                            mainError = ex;
                            this.ibState = JSDBGP.Constants.IB_STATE_START;
                        }
                        res = null;
                    }
                    if (mainError) {
                        this._doStderr(mainError.toString());
                    } else if (typeof res == "undefined" || res == null) {
                        // Do nothing
                    } else if (typeof res == "xml") {
                        this._doStdout(res.toXMLString());
                    } else {
                        this._doStdout(res.toString());
                    }
                }
            } else {
                // no decoded Data
                if (i_mode == -1 || this.ibState == JSDBGP.Constants.IB_STATE_START) {
                    // Nothing to do
                } else {
                    moreValue = 1;
                }
            }
        }
        response = this._get_xml_response_start_1();
        response.@status = this._getStopReason();
        response.@more = moreValue;
        response.@prompt = this.prompts[moreValue];
        this._addHeader_WriteXMLObject(response);
                            
    };

    this.onScriptInvalidated = function(tag) {
        delete this._knownScriptHandlers[tag];
        if (tag in this.console.pushedSources) {
            this.closePushedSource(tag);
        }
    };

    this._get_property_get_args = function(args, args_obj, info) {
        this._args_translate(args, args_obj, info);
        if (this.console.frames) {
            var frame_len = this.console.frames.length;
            if (args_obj.stackDepth >= frame_len) {
                args_obj.stackDepth = frame_len - 1;
            }
        } else {
            // ONly thing that makes sense.
            args_obj.stackDepth = 0;
            args_obj.context = this.propInfo.GlobalVars;
        }
        if (!args_obj.propertyLongName) {
            this.error_code = JSDBGP.Constants.E_INVALID_OPTIONS;
            throw new Error("Error in property args -- no propertyLongName specified");
        }
    };
    this._get_property_get_args._is_internal = true;

    this._re_accessor_part_start = /^([^.\[]+)(.*)$/;   // word before '.|['
    // match either leading dot, up to or or lb
    //              leading lb, digits, up to rb
    //              leading lb, double-quote..., skip bs... up to dq rb
    //              leading lb, single-quote..., skip bs... up to sq rb

    this._re_accessor_parts = [/^\.([^\[.]+)(.*)$/,      
                               /^\[([\d+])\](.*)$/,
                               /^\["((?:\\.|[^\\\"])+)"\](.*)$/,
                               /^\['((?:\\.|[^\\\'])+)'\](.*)$/
    ];
    this._parseAccessor = function(propertyLongName) {
        // For now assume it contains only dotted names
        var parts = [];
        var match = this._re_accessor_part_start.exec(propertyLongName);
        if (!match) {
            this.log.error("_parseAccessor: can't find accessors in "
                           + propertyLongName);
            // Graceful recovery on internal error
            return parts;
        }
        parts.push(match[1]);
        var rest = match[2];
        var parts_count = 1;
        while (rest.length > 0) {
            for (var i in this._re_accessor_parts) {
                var ptn = this._re_accessor_parts[i];
                match = ptn.exec(rest);
                if (match) {
                    parts.push(match[1]);
                    rest = match[2];
                    break;
                }
            }
            if (parts_count == parts.length) {
                this.log.error("Couldn't match an accessor in {" + rest + "}");
                // Graceful recovery on internal error
                return parts;
            }
            parts_count += 1;
        }
        return parts;
    };
    this._parseAccessor._is_internal = true;

    this._getProperty = function(currVal, propertyKey) {
        var result_obj = null;
        try {
            if (propertyKey in currVal) {



                result_obj = currVal[propertyKey];






            }
        } catch(ex4) {



        }
        return result_obj;
    };
    this._getProperty._is_internal = true;

    // This returns null if it failed, otherwise a result_obj:
    // isWrapped : true if it's a jsdIValue object, false if it's a native JS obj
    // obj : the value itself

    this._get_property_obj = function(args_obj) {
        var currVal;
        var result_obj;
        var bres;
        // Shuffle the args into locals for convenience
        var propertyLongName = args_obj.propertyLongName;
        var propertyKey = args_obj.propertyKey;
        var stackDepth = args_obj.stackDepth;
        var f;
        var context = args_obj.context;
        if (this.console.frames) {
            f = this._get_non_native_stack_frame(this.console.frames, stackDepth);
            if (!f) {






                return null;
            } else if (!f.isValid) {







                return null;
            }
    
            if (!this.getWalkedFrame(f, context)) {



                this.context_get({data: context + " " + stackDepth});
            }
        } else if (context != this.propInfo.GlobalVars) {
            this.error_code = JSDBGP.Constants.E_PROPERTY_NON_EXISTENT;
            throw new Error("Can't evaluate (" + propertyLongName + ")");
        } else {
            f = null;
        }

        if (context == this.propInfo.LocalVars) {
            var holdCurrVal = {};
            bres = f.eval(propertyLongName, this.dbgp_eval_ns, 1, holdCurrVal);
            if (typeof(bres) == "undefined") {










                this.error_code = JSDBGP.Constants.E_PROPERTY_NON_EXISTENT;
                throw new Error("Can't evaluate (" + propertyLongName + ")");
            }
            currVal = holdCurrVal.value.getWrappedValue();
            if (typeof currVal == "undefined"
                || (typeof currVal == "object"
                    && (currVal.name && currVal.name == "ReferenceError")
                    && (currVal.lineNumber && currVal.lineNumber == 1)
                    && (currVal.fileName && currVal.fileName == this.dbgp_eval_ns)
                    && (currVal.message && currVal.message.indexOf(" is not defined") > 0)
                    && (currVal.stack))) {







                return null;
            }
        } else {
            // No eval's needed -- just get the object and walk the
            // list of top-level accessors against it.
            
            if (context == this.propInfo.GlobalVars) {
                currVal = f ? (f.executionContext.isValid ? f.executionContext.globalObject.getWrappedValue() : null) : window;
            } else {
                currVal = f.thisValue.getWrappedValue();
            }
            var accessorList = this._parseAccessor(propertyLongName);
            var fqPropertyLongName = 'this.' + propertyLongName;
            for (var i in accessorList) {
                var newVal = this._getProperty(currVal, accessorList[i]);
                if (newVal) {
                    currVal = newVal;
                } else {






                    currVal = this.undefined;
                    return null;
                }
            }
        }
        if (propertyKey) {






            currVal = this._getProperty(currVal, propertyKey);
        }
        return currVal;
    };
    this._get_property_obj._is_internal = true;
    
    /*
     * property_get
     * Give full name (property_long_name), key (if any), stackDepth (def 0),
     *
     * and dump the results
     *
     * Syntax:
     * property_get -n propertyLongName [-c contextID] [-k propertyKey] [-d stackDepth]
     */

    this.property_get = function(args) {
        var args_obj = {};
        if (args.data) {
            this.error_code = JSDBGP.Constants.E_INVALID_OPTIONS;
            throw new Error("property_get -- specified data val of [" + args.data + "], but it won't be used");
        }
        this._get_property_get_args(args, args_obj,
                                    [['-n', 'propertyLongName'],
                                     ['-d', 'stackDepth', 'i', 0],
                                     ['-k', 'propertyKey'],
                                     ['-p', 'data_page', null, null],
                                     ['-c', 'context', 'i', this.propInfo.LocalVars],
                                     ]);
        var currVal = this._get_property_obj(args_obj);
        var fullName = args_obj.propertyLongName;
        var page_num = (args_obj.data_page == null) ? 0 : parseInt(args_obj.data_page);













        var hide_hidden = !this.settings.settings['show_hidden'][0];
        var max_data_size = this.settings.settings['max_data'][0];
        var outer_property =
        this.propInfo.expandTopLevelValue({name: fullName, value:currVal},
                                          hide_hidden,
                                          max_data_size
                                          );







        var amt_left = max_data_size - outer_property.toXMLString().length;
        if (outer_property.@children == 1) {
            var items_per_page = this.settings.settings['max_children'][0];
            outer_property.@page = page_num;
            outer_property.@page_size = items_per_page;



            var resa = this.propInfo.walkProperties(currVal, false, hide_hidden);
            if (outer_property.@numchildren != resa.length) {









                outer_property.@numchildren = resa.length;
            }
            var start_idx = page_num * items_per_page;
            var end_idx = (page_num + 1) * items_per_page; // 1 past end









            if (end_idx > resa.length) {
                end_idx = resa.length;
            }



            // resa is an array of [name, isWrapped, value] objects
            for (var i = start_idx; i < end_idx; i++) {
                if (!resa[i]) {



                    resa[i] = {name: "???", value:null, isWrapped:false};
                }
                var property = this.propInfo.getChildProperty(fullName,
                                                              resa[i].name,
                                                              resa[i].value,
                                                              amt_left,
                                                              hide_hidden);
                if (property) {



                    outer_property.property += property;
                } else {



                }
                if (amt_left >= 0) {
                    amt_left -= property.toXMLString().length;
                }
            }
        }
        var response = this._get_xml_response_start_1();
        if (args_obj.threadID != null) {
            response.@thread = args_obj.threadID;
        }
        response.property = outer_property;
        var resString = response.toXMLString();
        resString = resString.replace(/<property /g, "<property\n");
        this.console.session.writeData(this.xmlHeader + "\n" + resString);
        return false;
    };

    /*
     * property_set
     * Give full name (property_long_name), key (if any), stackDepth (def 0),
     * context (locals | this)
     * and value, plain at this point.
     *
     * and dump the results
     *
     * Syntax:
     * property_get -n propertyLongName [-c contextID] [-k propertyKey] [-d stackDepth]
     */
    
    this.property_set = function(args) {



        var args_obj = {};
       this._get_property_get_args(args, args_obj,
                                    [['-n', 'propertyLongName'],
                                     ['-d', 'stackDepth', 'i', 0],
                                     ['-k', 'propertyKey'],
                                     ['-p', 'data_page', null, null],
                                     ['-t', 'setter_data_type', null, null],
                                     ['-c', 'context', 'i', this.propInfo.LocalVars]
                                     // ,['-l', 'data_length', 'i', 0]
                                     ]);
        var success;
        var failureReason = null;
        try {
            if (!args.data) {
                this.error_code = JSDBGP.Constants.E_INVALID_OPTIONS;
                throw new Error("property_set -- failed to specify data val");
            }
            var newVal = this.utils.base64Decode(args.data);
            if (args_obj.setter_data_type == "string") {
                if (!this._stringEscape_re) {
                    this._stringEscape_re = /([\\\"\r\n])/g;
                    this._stringEscapeFunction = function(str) {
                        return "\\" + str;
                    };
                }
                newVal = ('"' +
                          newVal.replace(this._stringEscape_re,
                                         this._stringEscapeFunction)
                          + '"');
            }
            if (args_obj.propertyKey) {
                this.error_code = JSDBGP.Constants.E_INVALID_OPTIONS;
                throw new Error("Unexpected property_set request contained propertyKey " + args_obj.propertyKey);
            }
            var cmd = args_obj.propertyLongName + " = (" + newVal + ");";
            if (args_obj.context == this.propInfo.InstanceVars && cmd.indexOf("this.") != 0) {
                cmd = "this." + cmd;
            }
            var f;
            if (this.console.frames) {
                f = this._get_non_native_stack_frame(this.console.frames, args_obj.stackDepth);
            } else {
                f = null;
            }
            if (!f || !f.isValid) {



                eval(cmd);
                success = 1;
            } else {



                success = f.eval(cmd, this.dbgp_eval_ns, 1, {}) ? 1 : 0;
            }
        } catch(ex2) {
            this.log.exception(ex2);
            failureReason = ex2.name ? (ex2.name + ": ") : "";
            if (ex2.message) {
                failureReason += ex2.message;
            }
            success = 0;
        }
        var response = this._get_xml_response_start_1();
        response.@success = success;
        if (args_obj.threadID != null) {
            response.@thread = args_obj.threadID;
        }
        if (failureReason && failureReason.length > 0) {
            var message = <message />;
            // I think XML escapes the text...
            message.appendChild(new XML("[" + failureReason + "]"));
            var error = <error />;
            var ecode = JSDBGP.Constants.E_PROPERTY_NON_EXISTENT;
            error.@code = ecode[0];
            error.@apperr = ecode[1];
            error.message = message;
            response.error = error;
        }
        this._addHeader_WriteXMLObject(response);
        return false;
    };

    /* Write out objects as XML things
     *
     *
     * property_value -n propertyLongName [-c contextID] [-k propertyKey] [-d stackDepth]
     */

    this.property_value = function(args) {



        var args_obj = {};
        if (args.data) {
            this.error_code = JSDBGP.Constants.E_INVALID_OPTIONS;
            throw new Error("Error in property_value -- specified data val of [" + args.data + "], but it won't be used");
        }
        this._get_property_get_args(args, args_obj,
                                    [['-n', 'propertyLongName'],
                                     ['-d', 'stackDepth', 'i', 0],
                                     ['-k', 'propertyKey'],
                                     ['-c', 'context', 'i', this.propInfo.LocalVars],
                                     ]);
        var currVal = this._get_property_obj(args_obj);
        var fullName = args_obj.propertyLongName;
        var response = this._get_xml_response_start_1();
        var item = {name:fullName, isWrapped:false, value:currVal, stringify:true};
        var property = this.propInfo.expandTopLevelValue(item,
                                                         true, // hide_hidden
                                                         this.settings.settings['max_data'][0]
                                                         );
        response.property = property.text();
        response.@size = property.@size;
        response.@encoding = property.@encoding;
        if (args_obj.threadID != null) {
            response.@thread = args_obj.threadID;
        }
        this._addHeader_WriteXMLObject(response);
        return false;
    };

    this.run = function(args) {
        this.lastContinuationStatus = "break";
        this.lastTranID = this.transaction_id;
        this.stopReason = JSDBGP.Constants.STOP_REASON_RUNNING;
        this.mainObj.debuggerObj.setStopState(false);
        this.console.check_for_run_end_only = true;
        this.console.jsds.functionHook = this.console.callHook;
        this.doExitNestedEventLoop();
        return true;
    };

    this.extractInfoFrom_DBGP_URI = function(uri, retObj) {
        var match = this.utils.js_dbgp_javascript_url_re.exec(uri);
        if (match) {
            retObj.tag = match[1];
            retObj.uri = this.javascript_urls[retObj.tag];
            return true;
        }
        match = (this.utils.js_dbgp_with_base_re.exec(uri)
                 || this.utils.js_dbgp_no_base_re.exec(uri));
        if (match) {
            retObj.uri = match[1];
            retObj.tag = match[2];
            return true;
        }
        return false;
    };

    // For now replace spaces with encoded spaces.
    // Later write a general-purpose solution (if we need it).
    // Fix bug 65940
    this._fix_uri = function(uri) {
        return uri.replace(/ /g, "%20");
    };
    
    this.source = function(args) {
        var console = this.console;
        var beginLine = parseInt(this._getArg(args, '-b', -1));
        var endLine =   parseInt(this._getArg(args, '-e', -1));
        var fileURI =   this._getArg(args, '-f', '');

        fileURI = this._fix_uri(fileURI);
        
        var getExternalSource;
        var actualURI;
        var singleTag;
        var dbgp_info_obj = {};
        var resultObj = { finalSource : null, status : false };
        if (this.extractInfoFrom_DBGP_URI(fileURI, dbgp_info_obj)) {
            singleTag = dbgp_info_obj.tag;
            actualURI = dbgp_info_obj.uri;
            getExternalSource = false;
        } else {
            getExternalSource = true;
            if (fileURI.indexOf("dbgp:///") == 0) {
                actualURI = fileURI.substr("dbgp:///".length);
            } else {
                actualURI = fileURI;
            }
            singleTag = null;
        }
        var finalSource;
        var status;
        if (getExternalSource) {



            status = this.utils.getFileSource(actualURI, resultObj);
            if (!status) {





                var httpURI = this.mainObj.scripts.mapFileURI_to_HTTP_URI(actualURI);
                if (httpURI != actualURI) {
                    status = this.utils.getFileSource(httpURI, resultObj);




                }
            }
            if (status) {
                finalSource = resultObj.finalSource;
            } else {
                finalSource = (  "// File " + httpURI + " is unavailable.\n"
                                 + "\n"
                                 + "// Press the pretty-print button to see the function source.\n");
            }
        } else {
            finalSource = null;
        }
        if (typeof(finalSource) != "string" && getExternalSource) {
            if (finalSource) {






                finalSource = null;
            }
        }
        
        if (finalSource == null) {
            if (actualURI in console.scriptManagers) {
                var sm = console.scriptManagers[actualURI];
                var lines = [];
                var maxLineExtent = 0;
                if (singleTag) {
                    var jsdScript = sm.scriptWrappers[singleTag].jsdScript;
                    console.pushedSources[singleTag] = fileURI;
                    this.mainObj.breakpointManager.cloneBreakpoints(singleTag, jsdScript);
                    if (jsdScript.isValid) {
                        finalSource = jsdScript.functionSource;
                        if (actualURI.indexOf("javascript:") < 0) {
                            //Still a hack.  How did Ginda figure it out?
                            if (/^\s+function\b/.test(finalSource)) {
                                finalSource = "\n" + finalSource;
                            }
                        }






                    }
                } else {
                    for (var tag in sm.scriptWrappers) {
                        var sw = sm.scriptWrappers[tag];
                        var jsdScript = sw.jsdScript;
                        if (!jsdScript.isValid) {
                            continue;
                        } else if (jsdScript.functionName) {
                            // It's a script for a function without the
                            // function prototype, and will throw off line-numbering.
                            // Get the source from the top-level container.
                            continue;
                        }
                        var baseLineNumber = jsdScript.baseLineNumber;
                        var lineExtent = jsdScript.lineExtent;
                        if (maxLineExtent < lineExtent) {
                            maxLineExtent = lineExtent;
                        }
                        var functionSource = jsdScript.functionSource;



                        var functionLines = functionSource.split(/\n/);
                        for (var i = baseLineNumber; i < baseLineNumber + lineExtent; i++) {
                            if (typeof(lines[i - 1]) == "undefined") {
                                lines[i - 1] = functionLines[i];
                            }
                        }
                        for (var i = 0; i < maxLineExtent - 1; i++) {
                            if (typeof(lines[i]) == "undefined") {



                                lines[i] = "";
                            }
                        }
                    }
                    finalSource = lines.join("\n");
                }
            } else {
                this.error_code = JSDBGP.Constants.E_FILE_OPEN_FAILED;
                throw new Error("command.source -- couldn't find console.scriptManagers[" +
                                actualURI + "]");
            }
        }

        var response = this._get_xml_response_start_1();
        if (finalSource) {
            var finalLines = finalSource.split(/\r?\n/);
            if (beginLine == -1) beginLine = 0;
            if (endLine == -1) endLine = finalLines.length - 1;
            var culledLines = finalLines.slice(beginLine, endLine + 1);
            var culledSource = culledLines.join("\n");
            response.@success = 1;
            var encoding = 'base64';
            response.@encoding = encoding;
            var encVal = this.utils.base64Encode(culledSource);
            response.appendChild(encVal);
        } else {
            response.@success = 0;
        }
        var self = this;
        var prettyPrintStatus = singleTag ? 1 : 0;
        this.addNotification(function() {
            self.notifyPrettyPrint(actualURI, prettyPrintStatus);
        }, false);
        this._addHeader_WriteXMLObject(response);
        return false;
    };

    this._emitJSBanner = function() {
        var jsds = this.console.jsds;
        var s = ["JavaScript Interactive Shell",
                 "Using Mozilla JavaScript technology",
                 ("Version "
                  + jsds.implementationMajor + "."
                  + jsds.implementationMinor
                  + " ("
                  + jsds.implementationString
                  + ")"),
                 ""];
        this._doStdout(s.join("\n"));
    };

    this.status = function(args) {
        var response = this._get_xml_response_start_1();
        response.@status = this._getStopReason();
        // What it should be one day:
        // response.@status = this.console.startedAsInteractiveShell ? 'interactive' : this._getStopReason();
        this._addHeader_WriteXMLObject(response);
        if (this.in_interactive_shell && this._emitJSBanner) {
            var self = this;
            setTimeout(function() {
                self._emitJSBanner();
                self._emitJSBanner = null;
            }, 100);
        }
        return false;
    };

    this.wrapDoStatus = function(status) {
        if (this.lastContinuationCommand) {
            var response = this._get_xml_response_start_1();
            response.@command = this.lastContinuationCommand;
            response.@transaction_id = this.lastTranID;
            response.@status = status;
            response.@reason = "ok";
            this._addHeader_WriteXMLObject(response);
        }
    };
    this.wrapDoStatus._is_internal = true;

    this.stderr = function(args) {
        return this._common_stdio_redirection('stderr', args);
    };

    this.stdout = function(args) {
        var res = this._common_stdio_redirection('stdout', args);
        return res;
    };

    this.proxyerror = function(args) {
        alert("Debugger Proxy Error: "+this.utils.base64Decode(args.data));
        return false;
    };
    
    // Args not used
    this._common_stdio_redirection = function(stream_type, args) {
        var response = this._get_xml_response_start_1();
        response.@success = 1;
        this._addHeader_WriteXMLObject(response);
        return false;
    }
    this._common_stdio_redirection._is_internal = true;

    this.step_into = function(args) {
        this.console.check_for_run_end_only = true;
        return this._step_into_finish(args);
    }

    
    this._step_into_finish = function(args) {
        var console = this.console;
        if (console.frames) {
            this.mainObj.debuggerObj.setStopState(true);
            var topFrame = console.frames[0];
            if (topFrame.isValid) {
                console._stepPast = {
                    line : this.mainObj.debuggerObj.getCurrentLineByMode(topFrame),
                    frames_length : console.frames.length,
                    fileName : topFrame.script.fileName
                };
            } else {



            }
        }
        this.stopReason = JSDBGP.Constants.STOP_REASON_RUNNING;
        console.jsds.functionHook = console.callHook;
        this.doExitNestedEventLoop();
        return true;
    };

    this.step_out = function(args) {
        var console = this.console;
        if (console.frames.length == 1) {
            return this.run();
        }
        this.mainObj.debuggerObj.setStopState(false);
        console._stepOverDepth = console.frames.length - 1;
        console.jsds.functionHook = console.callHook;
        this.stopReason = JSDBGP.Constants.STOP_REASON_RUNNING;
        this.doExitNestedEventLoop();
        return true;
    };

    this.step_over = function(args) {
        this.console._stepOverDepth = this.console.frames.length;
        return this._step_into_finish(args);
    };

    this.stop = function(args) {
        // Mozilla/JS doesn't have a stop function, since it's embedded
        // in a UI.  Maybe throw an exception?



        if (this.console.initialized) {
            var startedProgram = this.stopReason != JSDBGP.Constants.STOP_REASON_STARTING;
            try {
                this.closeAllPushedSources(startedProgram);
            } catch(ex) {
                this.log.exception(ex);
            }
            if (startedProgram) {
                var response = this._get_xml_response_start_1();
                this.stopReason = JSDBGP.Constants.STOP_REASON_STOPPED;
                response.@status = this._getStopReason();
                this._addHeader_WriteXMLObject(response);
            }
            this.doExitNestedEventLoop();
            this.mainObj.debuggerObj.commence_shutdown();
        } else {



        }
        return true;
    };

    this.stack_depth = function(args) {
        // Walk the stacks looking for non-empty scripts.
        var response = this._get_xml_response_start_1();
        var num_active_frames = 0;
        var frames = this.console.frames;
        if (frames) {
            for (var i = 0; i < frames.length; i++) {
                var f = frames[i];
                if (f && f.isValid && f.script && f.script.fileName) {
                    num_active_frames += 1;
                } else {







                }
            }
        }
        response.@depth = num_active_frames;
        this._addHeader_WriteXMLObject(response);
        return false;
    };

    this._getScriptWrapper = function(jsdScript) {
        var tag = jsdScript.tag;
        var sm = this.mainObj.debuggerObj.internURL(jsdScript.fileName);
        if (!(tag in sm.scriptWrappers)) {
            sm.onScriptCreated(jsdScript);
        }
        return sm.scriptWrappers[tag];
    };

    // Return an array of :
    // 0: false: show the position as is, true: 
    // 1: true
    this._looksLikeScriptHandler = function(frame, jsdScript, url, functionName, frame_stack_length) {
        var pp = this.settings.settings['pretty_print'][0];
        if (pp) {
            return [true, null];
        } else if (url.indexOf("javascript:") == 0) {
            return [true, null];
        } else {
            var tag = jsdScript.tag;
            if (!(tag in this._knownScriptHandlers)) {
                var locationInfo = {};
                var res = this._looksLikeScriptHandler_aux(frame, jsdScript, url, functionName, frame_stack_length, locationInfo);
                this._knownScriptHandlers[tag] = [res, res ? locationInfo : null];
            }
            return this._knownScriptHandlers[tag];
        }
    };

    // memoize this info
    // Bug 65973: figure out what to do with the url arg
    this._looksLikeScriptHandler_aux = function(frame, jsdScript, url, functionName, frame_stack_length, locationInfo) {
        if (frame_stack_length == 1 && functionName in this._htmlEventHandlerNames) {
            var listValue = {value: null};
            var lengthValue = {value: 0};
            frame.scope.getProperties(listValue, lengthValue);
            if (lengthValue.value >= 1) {
                var prop1 = listValue.value[0];
                if (prop1.name.stringValue == "event" && (prop1.flags & prop1.FLAG_ARGUMENT) != 0) {
                    var inner_url;
                    var event = prop1.value.getWrappedValue();
                    if (event.currentTarget && event.currentTarget.baseURI) {
                        inner_url = event.currentTarget.baseURI;
                    } else if (jsdScript.fileName) {
                        inner_url = jsdScript.fileName;
                    } else {
                        this.log.debug("Couldn't get a filename");
                        return false;
                    }
                    this._populateLocationInfo(jsdScript, inner_url, locationInfo);
                    return true;
                }
            }
        }
        return false;
    };
    
    this._populateLocationInfo = function(jsdScript, url, locationInfo) {
        var possible_lines = [];
        var scriptFunctionName = jsdScript.functionName;
        var scriptFilename = jsdScript.fileName;







        var handler_line_number;
        var fileSourceLC = null;
        if (this.utils) {
            try {
                var resultObj = { finalSource : null, status : false };
                var status = this.utils.getFileSource(url, resultObj);
                if (status) {
                    fileSourceLC = resultObj.finalSource;
                }
            } catch(ex) {}
        }
        if (!fileSourceLC && typeof(RemoteFile) == "function") {
            var reader = new RemoteFile(url);
            try {
                reader.open();
                fileSourceLC = reader.read();
                reader = null;
            } catch(ex) {}
        }

        if (!fileSourceLC) {
            // Give up -- we can't get the raw source using the DOM,
            // as we'll be off by a few lines depending on how the
            // leading comments and XML declarations, HTML and HEAD
            // start tags are given.
            return false;
        }
        fileSourceLC = this._normalizeSearchableText(fileSourceLC);
        var scriptSourceLC = jsdScript.functionSource;
        var functionMatch = scriptSourceLC.match(/^\s+function /m);
        if (functionMatch) {
            // Skip this line
            var idx = scriptSourceLC.indexOf("\n");
            if (idx > 0) {
                scriptSourceLC = scriptSourceLC.substring(idx);
            }
            scriptSourceLC = this._normalizeSearchableText(scriptSourceLC);
        } else {
            return false;
        }
        var scriptLinesLC = scriptSourceLC.split("\n");
        var functionNameLC = jsdScript.functionName.toLowerCase();
        var lastPosn = 0;
        var handlerStartPosn;
        while (true) {
            var thisPosn = fileSourceLC.indexOf(functionNameLC, lastPosn);
            if (thisPosn == -1) {



                break;
            }
            handlerStartPosn = thisPosn;

            // Handlers can't contain a less-than char, so that's a
            // definite limit on how far we can search.
            var tagEnd = fileSourceLC.indexOf("<", thisPosn);
            var constrainedSource = fileSourceLC.substring(thisPosn, tagEnd);
            var lastPosn2 = 0;
            var thisPosn2;
            var failed = false;
            for (var lineIdx = 0; lineIdx < scriptLinesLC.length; lineIdx++) {
                var scriptLine = this._strip(scriptLinesLC[lineIdx]);
                if (scriptLine.length == 0) {
                    continue;
                }
                // Semi-colons are optional at end of each line
                scriptLine = scriptLine.replace(/;\s*$/, '');
                if (scriptLine == "}" && lineIdx == scriptLinesLC.length - 1) {



                    break;
                }








                thisPosn2 = constrainedSource.indexOf(scriptLine, lastPosn2);
                if (thisPosn2 == -1) {
                    // Not here



                    failed = true;
                    break;
                } else {
                    lastPosn2 = thisPosn2 + 1;
                }
            }
            if (!failed) {
                // Grab the first one.



                possible_lines.push(this._countNewlines(fileSourceLC.substr(0, handlerStartPosn)) + 1);






            }
            lastPosn = thisPosn + constrainedSource.length;
        }
        switch (possible_lines.length) {
        case 1:
            locationInfo.lineno = possible_lines[0];
            locationInfo.fileName = url;
            return true;
        case 0:




        default:



            return false;
        }
    };

    // Pull out all quotes and backslashes that get normalized
    // differently.  This stuff is so error-prone in general...
    this._normalizeSearchableText = function(s) {
        var s1 = s.replace(/[\'\"\\ \t]/g, '').toLowerCase();
        return s1;
    };

    this._strip = function(s) {
        return s.replace(/^\s+/,"").replace(/\s+$/,"");
    };

    this._countNewlines = function(s) {
        return s.replace(/[^\n]/g,'').length;
    };

    this.stack_get = function(args) {
        var response = this._get_xml_response_start_1();
        var frames = this.console.frames;
        if (!frames) {



            this._addHeader_WriteXMLObject(response);
            return;
        }
        var stackDepth = this._getArg(args, '-d', 0);
        if (stackDepth < 0) {
            this.error_code = JSDBGP.Constants.E_STACK_DEPTH_INVALID;
            throw new Error("stack_get: Got a negative stack_get of " + stack_get);
        } else if (stackDepth >= frames.length) {




            stackDepth = frames.length - 1;
        }
        for (var i = stackDepth; i < frames.length; i++) {
            var f = frames[i];
            if (f.isNative || !f.isValid) {
                continue;
            }
            var jsdScript = f.script;
            if (!jsdScript) {
                // This frame is native, so skip it
                this.log.error("stack_get: stack frame " + i + "isn't native, but has no script!");
                continue;
            } else if (!jsdScript.isValid) {
                // This frame is native, so skip it
                this.log.error("stack_get: stack frame " + i + "'s script no longer valid");
                continue;
            }
            var functionName = f.functionName || "<anonymous>";
            var fileName;
            var line = null;
            // XXX sometimes f.script does not exist
            var mapMode = Components.interfaces.jsdIScript.PCMAP_SOURCETEXT;
            if (jsdScript.fileName) {
                var fileName_raw = this.utils.canonicalize_uri(jsdScript.fileName);







                var locationInfoTuple = this._looksLikeScriptHandler(f, jsdScript, fileName_raw, functionName, frames.length);
                if (locationInfoTuple[0]) {
                    var locationInfo = locationInfoTuple[1];
                    if (locationInfo && locationInfo.lineno) {
                        fileName = this.mainObj.scripts.mapHTTP_URI_to_FileURI(locationInfo.fileName);
                        line = locationInfo.lineno;
                    } else {
                        var sw = this._getScriptWrapper(jsdScript);
                        // assert sw != null
                        if ('_dbgp' in sw) {
                            fileName = sw._dbgp;
                        } else {
                            if (!('_appid' in this) || (this._appid == 0)) {
                                // We couldn't get the PID, so set it to the time
                                // to the nearest .1 seconds over the last
                                // 115 days.  Along with the script tag this
                                // should give a small but unique number.
                                // The purpose of doing this is to prevent
                                // collisions in long-running instances of Komodo.
                            
                                this._appid = Math.round((new Date().valueOf() % 1e7) / 100);
                            }
                            if (!('_baseName_re' in this)) {
                                this._baseName_re = /.*\/([^\/.]+)(?:\.[^\/]*)?$/;
                            }
                            var baseFileName;
                            if (fileName_raw.indexOf("javascript:") == 0) {
                                this.javascript_urls[jsdScript.tag] = fileName_raw;
                                fileName = ("dbgp://js/"
                                            + this._appid
                                            + "/"
                                            + "<javascript-url>"
                                            + ":"
                                            + jsdScript.tag
                                            + ".js");
                            } else {
                                var matchObj = this._baseName_re.exec(fileName_raw);
                                if (matchObj) {
                                    fileName = ("dbgp://js/"
                                                + fileName_raw
                                                + "/"
                                                + this._appid
                                                + "/"
                                                + matchObj[1]
                                                + ":"
                                                + jsdScript.tag
                                                + ".js");
                                } else {
                                    fileName = ("dbgp://js/"
                                                + this._appid
                                                + "/"
                                                + fileName_raw
                                                + "<index>:"
                                                + jsdScript.tag
                                                + ".js");
                                }
                            }
                            sw._dbgp = fileName;
                        }
                        mapMode = Components.interfaces.jsdIScript.PCMAP_PRETTYPRINT;
                    }
                } else {
                    fileName = this.mainObj.scripts.mapHTTP_URI_to_FileURI(fileName_raw);
                }
            } else {
                this.log.error("stack_get: stack frame " + i + "has a script but no filename, skipping...");
                continue;
            }
            if (line == null) {




                var pc = f.pc;



                try {
                    line = jsdScript.pcToLine(pc, mapMode);
                } catch(ex) {
                    line = 0;
                }
            }
            var stack = <stack />;
            stack.@level = i;
            stack.@type = "file";
            stack.@filename = fileName;
            stack.@lineno = line;
            stack.@where = functionName;
            response.stack += stack;
        }
        this._addHeader_WriteXMLObject(response);
    };

    // Returns an array of [name, [raw-value, printable-value]] pairs.

    this._exclusion_filter = function(name, val, exclusions) {
        if (!exclusions) return false;
        for (var k in exclusions) {
            if (name.indexOf(k) >= 0) {
                if (!exclusions[k]) return true;
                else if (typeof(val) == exclusions[k]) {
                    return true;
                }
            }
        }
        return false;
    };

    this.typemap_get = function(args) {
        var response = this._get_xml_response_start_1();
        var xsdNS = new Namespace('xsd', "http://www.w3.org/2001/XMLSchema");
        var xsiNS = new Namespace('xsi', "http://www.w3.org/2001/XMLSchema-instance");
        response.addNamespace(xsdNS);
        response.addNamespace(xsiNS);
        // Schema, CommonTypeName (type attr) LanguageTypeName (name attr)
        var names = [['boolean', 'bool'],
                     ['float'],
                     ['integer', 'int'],
                     ['string']];
        for (var i in names) {
            var e = names[i];
            var xsdName = e[0];
            var commonTypeName = (e.length >= 2 ? e[1] : xsdName);
            var languageTypeName = (e.length >= 3 ? e[2] : commonTypeName);
            response.map += <map />;
            response.map[i].@type = commonTypeName;
            response.map[i].@name = languageTypeName;
            response.map[i].@xsiNS::type = "xsd:" + xsdName;
        }
        this._addHeader_WriteXMLObject(response);
        return false;
    };

    this._getStopReason = function() {
        return this.stopReasons[this.stopReason];
    };
    this._getStopReason._is_internal = true;

    this._get_xml_output = function(streamName) {
        var response = <stream />;
        response.@xmlns = this.dbgpNamespace;
        response.@type = streamName;
        response.@encoding = this.settings.settings['data_encoding'][0];
        return response;
    };

    this._get_xml_response_start_1 = function() {
        var response = <response />;
        response.@xmlns = this.dbgpNamespace;
        response.@command = this.cmd;
        response.@transaction_id = this.transaction_id;
        if (this.thread_id) {
            response.@thread_id = this.thread_id;
        }
        return response;
    };
    this._get_xml_response_start_1._is_internal = true;

    this._get_xml_notify_start_1 = function() {
        var notification = <notify />;
        notification.@xmlns = this.dbgpNamespace;
        return notification;
    };
    this._get_xml_notify_start_1._is_internal = true;

    this._getArg = function(args, argName, defaultValue) {
        if (argName in args) return args[argName];
        if (arguments.length >= 3) return defaultValue;
        return null;
    };
    this._getArg._is_internal = true;

    this.sendInitString = function(stopReason) {
        var idekey = null;
        var env = new this.Environment(this.utils);
        
        // retrieve proxy configuration data
        var oPrefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("");
        var use_proxy = oPrefs.getBoolPref("jsdbgp.useProxy");
        if (use_proxy) {
            idekey = oPrefs.getCharPref("jsdbgp.proxyName");
        // XXX todo: if use_proxy and no proxy_name, get proxy_name from env: USER or USERNAME
        }
        if (use_proxy && !idekey) {
            idekey = env.get('USER', 0);
            if (!idekey)
                idekey = env.get('USERNAME', 0);
            if (!idekey) {
                this.error_code = JSDBGP.Constants.E_UNKNOWN;
                throw new Error("DBGP using proxy but no key is available");
            }
        }
        var response = <init />;
        response.@xmlns = this.dbgpNamespace;
        try {
            this._appid = Components.classes["@mozilla.org/process/util;1"].createInstance(Components.interfaces.nsIProcess).pid;
        } catch(ex1) {
            this._appid = env ? env.get('DEBUGGER_PID', 0) : 0;
        }
        response.@appid = this._appid;
        if (use_proxy && idekey != null) {
            response.@idekey = idekey;
        }
        
        response.@thread = "main 0";
        response.@language="JavaScript";
        response.@protocol_version="1";
        if (this.in_interactive_shell) {
            response.@interactive=">";
        } else {
            response.@fileuri = this.console.target_file_name;
        }
        if (this.session_key) {



            response.@session = this.session_key;
        }
            
        this._addHeader_WriteXMLObject(response);



        this.stopReason = stopReason;
        this.lastContinuationCommand = null;
        this.lastContinuationStatus = "break";
        this.lastTranID = 0;
    };

    // Writer routines

    this.xmlHeader = '<?xml version="1.0" encoding="utf-8" ?>';
    this.dbgpNamespace = "urn:debugger_protocol_v1";

    // Code to track whether we've walked a given frame
    
    this.makeWalkedFrameTag = function(f, type) {
        return (f.script.tag.toString()
                + ":"
                + f.pc
                + ":"
                + type);
    };

    this.setWalkedFrame = function(f, type) {
        this.console.walkedFrames[this.makeWalkedFrameTag(f, type)] = true;
    };

    this.getWalkedFrame = function(f, type) {
        return this.console.walkedFrames[this.makeWalkedFrameTag(f, type)];
    };

    this.closeAllPushedSources = function(startedProgram) {
        for (var tag in this.mainObj.console.pushedSources) {
            this.closePushedSource(tag, startedProgram);
        }
    };

    this.closePushedSource = function(tag, startedProgram) {
        var url = this.mainObj.console.pushedSources[tag];
        if (url) {
            this.mainObj.bpsMediator.doScriptDeleted(tag);
            if (startedProgram) {
                var self = this;
                this.addNotification(function() {
                    self.notifyScriptDeleted(url);
                }, true);
            }
            delete this.mainObj.console.pushedSources[tag];
        }
    };

    // Notifications
    this.addNotification = function(func, runImmediately) {
        if (runImmediately) {
            try {
                func();
            } catch(ex) {
                this.log.exception(ex);
            }
        } else {
            this.notifications.push(func);
        }
    };

    this._doNotifyBreakpointInfo = function(name, bkpt_id) {
        var notification = this._get_xml_notify_start_1();
        notification.@name = name;
        var bkpt = this.mainObj.breakpointManager.getBreakpoint(bkpt_id);
        if (!bkpt) {



            return;
        }
        var breakpoint = bkpt.display_e4x();
        breakpoint.@id = bkpt_id;
        notification.breakpoint = breakpoint;
        this._addHeader_WriteXMLObject(notification);
    };

    this.notifyBreakpointSet = function(bkpt_id) {
        try {
            this._doNotifyBreakpointInfo('breakpoint_set', bkpt_id);
        } catch(ex) {
            this.log.exception(ex);
        }
    };

    this.notifyBreakpointDelete = function(bkpt_id) {
        try {
            var notification = this._get_xml_notify_start_1();
            notification.@name = 'breakpoint_delete';
            notification.appendChild(bkpt_id.toString());
            this._addHeader_WriteXMLObject(notification);
        } catch(ex) {
            this.log.exception(ex);
        }
    };

    this.notifyBreakpointModify = function(bkpt_id) {
        try {
            this._doNotifyBreakpointInfo('breakpoint_update', bkpt_id);
        } catch(ex) {
            this.log.exception(ex);
        }
    };

    this.notifyPrettyPrint = function(url, prettyPrintStatus) {
        var notification = this._get_xml_notify_start_1();
        notification.@name = 'pretty_print';
        notification.appendChild(prettyPrintStatus ? "1" : "0");
        this._addHeader_WriteXMLObject(notification);
        this.settings.settings['pretty_print'][0] = prettyPrintStatus ? 1 : 0;
    };

    this.notifyScriptDeleted = function(url) {
        var notification = this._get_xml_notify_start_1();
        notification.@name = 'script_destroyed';
        notification.appendChild(url);
        this._addHeader_WriteXMLObject(notification);
    };

    this.notifyScriptCreated = function(url) {
        var notification = this._get_xml_notify_start_1();
        notification.@name = 'script_destroyed';
        notification.appendChild(url);
        this._addHeader_WriteXMLObject(notification);
    };

    // Moz Environment wrapper

    this.Environment = function(utils) {
        this.mozEnv = Components.classes["@mozilla.org/process/environment;1"].getService(Components.interfaces.nsIEnvironment);
    };

    this.Environment.prototype.get = function(name, defaultValue) {
        if (this.mozEnv.exists(name)) {
            return this.mozEnv.get(name);
        } else {
            return defaultValue;
        }
    };

    this.Environment.prototype.exists = function(str) {
        return this.mozEnv.exists();
    };

    this.Environment.prototype.set = function(name, val) {
        this.mozEnv.set(name, val);
    };
}
