/* -*- Mode: JavaScript; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * Copyright (c) 2006, 2007 ActiveState Software Inc.
 * Refer to the license in the Komodo IDE installation
 * for the licensing information of this source code.
 *
 * winger-static.js -- where the debugger starts -- globals defined here.
 */

if (typeof(JSDBGP) == 'undefined') {
    var JSDBGP = {};
}

JSDBGP.AppCheck = {
    FIREFOX_ID : "{ec8030f7-c20a-464f-9b0e-13a3a9e97384}",
    
    getAppID : function() {
        var id;
        var appInfoCID = "@mozilla.org/xre/app-info;1";
        if (appInfoCID in Components.classes) {
            // running under Mozilla 1.8 or later
            id = Components.classes[appInfoCID].getService(Components.interfaces.nsIXULAppInfo).ID;
        } else {
            id = Components.classes["@mozilla.org/preferences-service;1"]
            .getService(Components.interfaces.nsIPrefBranch)
            .getCharPref("app.id");
        }
        return id;
    },
    isXULApp: function() {
        // XXX FIXME need a way to debug firefox itself, but this allows
        // for debugging other xul apps for now
        return JSDBGP.AppCheck.getAppID() != JSDBGP.AppCheck.FIREFOX_ID;
    },
    __end__ : null
};

// Almost all the objects in this app are singletons,
// so there's no need for prototypes.  All data hangs
// off the JSDBGP.main object, and other classes are
// defined within the JSDBGP pseudo-namespace object.

// scriptManagers and scriptWrappers need to be kept around across all debugging
// sessions, because doing a start() and stop() won't clear them.

JSDBGP.MainClass = function() {
    this.console = {
        initialized : false,
        breakpoints : {},
        lastWindow : null,
        nested_event_loop_depth : 0,
        session : null,
        target_file_name : null,
        check_for_run_end_only : false,
        isXulApp: JSDBGP.AppCheck.isXULApp(),
        scriptManagers : {}, // hash on url
        scriptWrappers : {}, // hash tag => JSDBGP.Scripts.ScriptWrapper
        pushedSources : {}, // hash tag => URI
        __end__ : null
    };
    this.log = JSDBGP.Logging.getLogger("winger.main");
    this.log.setLevel(JSDBGP.Logging.DEBUG);




    this.properties = new JSDBGP.Properties(this);
    this.bpsMediator = new JSDBGP.BreakpointScriptMediator(this);
    this.scripts = new JSDBGP.Scripts(this);
    this.utils = new JSDBGP.Utils(this);
    this.breakpointManager = new JSDBGP.BreakpointManager(this);

    // 'debugger' is a reserved word in JS, and
    // JS is broken in the way it doesn't allow reserved words after
    // dots.
    this.debuggerObj = new JSDBGP.Debugger(this);
    this.settings = new JSDBGP.Settings();
    this.console.command = new JSDBGP.Command(this);

    // Call these routines so objects that hang off the main object
    // can reference one another.
    this.breakpointManager.finalInit();
    this.console.command.finalInit();

    this.init = function(host_name, host_port, session_key, is_js_shell) {
        this.console._lastStackDepth = -1;
        this.debuggerObj.initDebugger(this, host_name, host_port,
                                      session_key, is_js_shell);

        this.settings.init();
        this.settings.completeSettings(this.console.jsds);
        this.console.initialized = true;
        
        if (this.console.target_file_name.lastIndexOf(".js")
            == this.console.target_file_name.length - 3) {

            var subscript_loader = Components.classes["@mozilla.org/moz/jssubscript-loader;1"].getService(Components.interfaces.mozIJSSubScriptLoader);
        
            var fname = this.console.target_file_name;



            try {
                // Have the script loaded into an empty object.
                var loader = { doit : subscript_loader.loadSubScript };
                loader.doit(fname);



            } catch(ex) {
                this.log.exception(ex, "wrapInitDebugger => ");
            }
            // Craft our own stop command here
            this.console.command.wrapDoStatus("stopped");
            this.debuggerObj.detachDebugger();
        } else if (is_js_shell) {
            // Nothing to do here.
        } else {
            this.debuggerObj.setStopState(true);
        }
    };

// From http://developer.mozilla.org/en/docs/Using_nsIXULAppInfo

    this._getCurrentURI = function() {
        var currentURI = window.location.href;
        try {
            // assume we're firefox, otherwise, fallback to window location
            currentURI = gBrowser.mCurrentBrowser.currentURI.spec;
        } catch(ex) {



        }



        return currentURI;
    };

    this.doStart = function(__target_file_name, host_name, host_port,
                            session_key, is_js_shell) {
        var target_file_name = this._getCurrentURI();
        if (!target_file_name)
            throw new Error("Target file is not defined");
        if (!this.console.initialized) {
            this.log.debug("About to initialize the Komodo JavaScript Debugger");
            this.console.target_file_name = target_file_name;
            
            // We need to track the parent of any documents we try to debug
            // with chrome:// URI schemes, so we don't step into internal
            // browser code.
            var target_url = (Components.classes["@mozilla.org/network/io-service;1"]
                              .getService(Components.interfaces.nsIIOService)).newURI(target_file_name, null, null).spec;
            var idx = target_url.indexOf("?");
            if (idx >= 0) {
                target_url = target_url.substring(0, idx);
            }
            idx = target_url.lastIndexOf("/");
            if (idx >= 0) {
                this.console.parent_dir = target_url.substring(0, idx);
            } else {
                this.console.parent_dir = target_url;
            }
            
            this.console.canon_target_file_name = this.utils.canonicalize_uri(target_file_name);
            this.init(host_name, host_port, session_key, is_js_shell);
            return true;
        } else {



        }
        return false;
    };
    
    this.doStop = function() {
        if (this.console.initialized) {
            this.log.debug("stopping and detaching debugger...");
            this.console.command.dispatch('stop');
            // console.command.stop => debuggerObj.commence_shutdown()
            // => jsds.off() and detachDebugger()
            //
            // this.debuggerObj.detachDebugger();
        }
    };

    this.doBreak = function() {
        var fn = function() {
            var enabled = document.getElementById('Tasks:JSDBGPBreak');
            if (enabled && enabled.getAttribute('checked') == 'true')
                enabled.setAttribute('checked', 'false');
        }
        if (this.console.jsds) {
            this.console.command.doBreakNextStatement(fn);
        }
    }

    // Callbacks from this.init ==> debuggerObj.initDebugger => new Session

    this.showConnected = function() {
        JSDBGP.ui_set_status(JSDBGP.Constants.DBGP_STATE_CONNECTED);
        document.getElementById('Tasks:JSDBGPEnable').setAttribute('checked','true');
        document.getElementById('Tasks:JSDBGPBreak').removeAttribute('disabled');
    };

    this.showDisconnected = function() {
        this.console.initialized = false;
        JSDBGP.winger_reset_ui(JSDBGP.Constants.DBGP_STATE_DISCONNECTED);
    };
    
    this.doDisconnect = function() {
        this.debuggerObj.detachDebugger();
        this.showDisconnected();
    }

};

//XXX: get two more icons

JSDBGP.status_to_file = {
    // JSDBGP.Constants.
    'DBGP connecting...' : 'status_connected.png',
    'DBGP Connected' : 'status_connected.png',
    'DBGP state: run' : 'status_connected.png',
    'DBGP state: break' : 'status_connected_break.png',
    'DBGP Disconnected' : 'status_post_connect.png',
    'DBGP Unused' : 'status_pre_connect.png'
};


JSDBGP.ui_set_status = function (textCode) {
    try {
        var statusElement = document.getElementById('jsdbgp-status');
        statusElement.setAttribute('tooltiptext',
                     (JSDBGP.Constants.DBGB_STATE_PREFIX
                      + JSDBGP.Constants.DBGB_STATE_STRINGS[textCode]));
        var icon_path = statusElement.getAttribute('src');
        var new_base = JSDBGP.Constants.DBGB_STATE_ICONS[textCode];
        if (new_base) {
            var parts = icon_path.split('/');
            parts[parts.length - 1] = new_base;
            var new_path = parts.join('/');
            if (new_path != icon_path) {
                statusElement.setAttribute('src', new_path);
            } else {



            }
        } else {





        }
    } catch(ex) {



    } 
}

JSDBGP.winger_reset_ui = function(textCode) {
    document.getElementById('Tasks:JSDBGPEnable').setAttribute('checked','false');
    document.getElementById('Tasks:JSDBGPBreak').setAttribute('checked','false');
    document.getElementById('Tasks:JSDBGPBreak').setAttribute('disabled', 'true');
    JSDBGP.ui_set_status(textCode);
}

// Other handlers
JSDBGP.winger_enable = function () {
    var enabled = document.getElementById('Tasks:JSDBGPEnable');
    var checked = JSDBGP._winger_button_on(enabled);
    if (!checked) {
        JSDBGP.winger_start();
    } else {
        JSDBGP.winger_stop();
    }
}


JSDBGP._complain_no_jslib = function() {
    alert("The ActiveState JavaScript debugger requires jslib.\n"
          + "\n"
          + "We include jslib.xpi under the Komodo install directory. "
          + "Alternately, you can use the JavaScript preferences "
          + "panel in Komodo to install everything required to "
          + "debug JavaScript in Komodo.");
    return false; // return same type of val as JSDBGP.main.doStart
}

JSDBGP.winger_command = function (param, is_js_shell) {
    var params = param.split(':');
    var host_name = params[0];
    var host_port = params[1];
    var session_key = (params.length > 2) ? params[2] : null;
    if (typeof(JS_LIB_LOADED) == "undefined") {
        return this._complain_no_jslib();
    }
    try {
        if (typeof(JSDBGP.main) == 'undefined') {
            JSDBGP.main = new JSDBGP.MainClass();
        }
        return JSDBGP.main.doStart(null, host_name, host_port, session_key, is_js_shell);
    } catch(ex) {
        dump(ex.toString() + "\n");
        return false;
    }
}

//force update
JSDBGP.winger_start = function () {
    if (typeof(JS_LIB_LOADED) == "undefined") {
        return this._complain_no_jslib();
    }
    /* XXX leave out till I finish
    try {
        if (!("content_policy" in JSDBGP)) {
            dump("creating JSDBGP policy handler...\n");
            JSDBGP.content_policy = Components.classes["@activestate.com/jsDBGPPolicy;1"].createInstance(Components.interfaces.nsIContentPolicy);
            dump("...JSDBGP policy handler created\n");
        }
    } catch(e) {
        dump(e+"\n");
    }
    */
    try {
        if (typeof(JSDBGP.main) == 'undefined') {
            JSDBGP.main = new JSDBGP.MainClass();
        }
        var product_name = "Komodo JavaScript Debugger"
        var heading = product_name + ":\n";
        var pref_locn = "the Komodo JavaScript Debugger Options";
        
        var oPrefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("");
        if (!oPrefs) {
            alert(heading +
                  "Internal error: the debugger can't find Firefox Preferences.\n" +
                  "Please contact support@activestate.com");
            return false;
        }
        // dbgp configuration
        var host_name = oPrefs.getCharPref("jsdbgp.host");
        if (!host_name) {
            alert(heading +
                  "Host Name is not set in " + pref_locn);
            return false;
        }
        var host_port = oPrefs.getIntPref("jsdbgp.port");
        if (typeof(host_port) != "number") {
            alert(heading +
                  "Host Port is not set in " + pref_locn);
            return false;
        }
        /* Not used yet
        var use_proxy;
        try {
            use_proxy = oPrefs.getIntPref("jsdbgp.useProxy");
            if (typeof(use_proxy) == "undefined") {
                use_proxy = null;
            }
        } catch {
            use_proxy = null;
        }
        */
        var session_key;
        try {
            session_key = oPrefs.getIntPref("jsdbgp.proxyName");
            if (typeof(session_key) == "undefined") {
                session_key = null;
            }
        } catch(ex) {
            session_key = null;
        }
        // XXX need to get the primary script to debug and pass to doStart
        // This will be needed for chrome debugging.
        return JSDBGP.main.doStart(null, host_name, host_port, session_key, false);
    } catch(ex) {
        dump(ex.toString() + "\n");
        return false;
    }
}


JSDBGP.winger_stop = function () {
    JSDBGP.main.doStop();
    JSDBGP.winger_reset_ui(JSDBGP.Constants.DBGP_STATE_DISCONNECTED);
}

JSDBGP._winger_button_on = function(button) {
    var buttonObj = typeof(button) == "string" ? document.getElementById(button) : button;
    return buttonObj.getAttribute('checked') && buttonObj.getAttribute('checked') != 'false';
}

JSDBGP.winger_break = function () {
    if (JSDBGP._winger_button_on('Tasks:JSDBGPEnable')) {
        var break_button = document.getElementById('Tasks:JSDBGPBreak');
        if (!JSDBGP._winger_button_on(break_button)) {
            break_button.setAttribute('checked','true');
            return JSDBGP.main.doBreak();
        } else {
            break_button.setAttribute('checked','false');
            // Update the controller
            if (JSDBGP.main && JSDBGP.main.debuggerObj) {
                JSDBGP.main.debuggerObj.setStopState(false);
            }
        }
    }
    return false;
}

// External API for other extensions:
// bool JSDBGP.isDebuggerEnabled();
// string JSDBGP.version;

JSDBGP.isDebuggerEnabled = function () {
    if (!JSDBGP.main) {
        return false;
    } else {
        return JSDBGP.main.isDebuggerEnabled();
    }
};

JSDBGP.__defineGetter__('version', function () {
    try {
        var bundle = document.getElementById("bundle_jsdbgp");
        if (!bundle) {
            dump("JSDBGP.version: id(bundle_jsdbgp) is null\n");
            return null;
        }
        return bundle.getString("jsdbgp_version");
    } catch(ex) {
    }
    return null;
});

JSDBGP.MainClass.prototype.isDebuggerEnabled = function() {
    return this.console.initialized;
};

JSDBGP.load = function() {
    // Determine if this is a first-run for this version of the extension.
    var prefSvc = Components.classes["@mozilla.org/preferences-service;1"]
            .getService(Components.interfaces.nsIPrefBranch);
    var currentVersion = JSDBGP.version;
    var previousVersion = null;
    var isFirstRun = false;
    try {
        var previousVersion = prefSvc.getCharPref("jsdbgp.version");
        if (previousVersion != currentVersion){
            prefSvc.setCharPref("jsdbgp.version", currentVersion);
            isFirstRun = true;
        }
    } catch(ex) {
        prefSvc.setCharPref("jsdbgp.version", currentVersion);
        isFirstRun = true;
    }
    //alert("isFirstRun: "+isFirstRun);

    if (isFirstRun && typeof(JSDBGP.firstRun)!='undefined') {
        // firstRun is application specific, and is implemented in xx-overlay.xul
        JSDBGP.firstRun();
    }
}


JSDBGP.unload = function() {
}

JSDBGP.launch_tab = function(aUrl) {
    var browser = document.getElementById("content");
    var tab = browser.addTab(aUrl);  
    browser.selectedTab = tab;
};

JSDBGP.show_welcome_page = function() {
    setTimeout(JSDBGP.launch_tab, 1000,
               "chrome://jsdbgp/content/welcome/welcome.html");
}

window.addEventListener("load", JSDBGP.load, false);
window.addEventListener("unload", JSDBGP.unload, false);

